package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.DesiredStatus;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecutionFinishedCallback;
import be.iminds.ilabt.jfed.highlevel.jobs.Job;
import be.iminds.ilabt.jfed.highlevel.jobs.State;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.ListSliceResourcesTask;
import be.iminds.ilabt.jfed.highlevel.tasks.StatusTask;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import com.google.common.base.Strings;
import com.google.common.collect.Iterables;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.Duration;
import java.time.Instant;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

/**
 * This state consists out of two loops:
 * <p/>
 * - Inner loop does a StatusTask, and repeats it until it reaches a final state
 * - Outer loop will perform an extra ListResources when GetStatus reports ok, because sometimes the status changes to a non-final state when querying ListResources
 * <p/>
 * We repeat both loops until both Status and ListResources report a final state
 * <p/>
 * <p/>
 * User: twalcari
 * Date: 11/27/2015
 * Time: 17:08
 */
public class UpdateUntilFinalOpStateTaskState extends State {
    private static final Logger LOG = LoggerFactory.getLogger(UpdateUntilFinalOpStateTaskState.class);
    private static final Duration DEFAULT_CHECK_READY_INTERVAL = Duration.ofSeconds(10);
    private static final int MAX_SOFT_FAIL_COUNT = 5;
    private static final int MAX_RESOURCES_FAILED_RETRY_COUNT = 5;
    private static final int MAX_BUSY_COUNT = 30 * 6; //30 minutes

    @Nonnull private final Job<?> job;
    @Nonnull private final SfaExperimentPart experimentPart;
    @Nullable private final Duration maxWaitUntilReady;

    @Nonnull private final Experiment experiment;
    @Nonnull private final HighLevelTaskFactory hltf;
    @Nonnull private Duration checkReadyInterval;

    private Instant timeoutTime;
    private int softFailCount = 0;
    private int resourcesFailedRetryCount = 0;
    private int busyCount = 0;
    private int triesCount = 0;
    private final StatusTask statusTask;
    private boolean waitForFirstCheck;

    private @Nonnull DesiredStatus desiredStatus;
    private boolean hadTimeout;

    protected UpdateUntilFinalOpStateTaskState(@Nonnull Job<?> job,
                                               @Nonnull SfaExperimentPart experimentPart,
                                               @Nullable Duration checkReadyInterval,
                                               @Nullable Duration maxWaitUntilReady,
                                               @Nonnull HighLevelTaskFactory hltf,
                                               @Nonnull DesiredStatus desiredStatus) {
        super(String.format("Waiting for resources on %s to become ready", experimentPart.getName()));
        this.job = job;
        this.experimentPart = experimentPart;
        this.checkReadyInterval = checkReadyInterval == null ? DEFAULT_CHECK_READY_INTERVAL : checkReadyInterval;
        this.maxWaitUntilReady = maxWaitUntilReady;
        this.hltf = hltf;
        this.experiment = experimentPart.getExperiment();
        this.desiredStatus = desiredStatus;
        this.hadTimeout = false;

        statusTask = hltf.status(experiment.getSlice(), experimentPart.getConnectSfaAuthority());
    }

    private void registerStart() {
        if (maxWaitUntilReady != null)
            timeoutTime = Instant.now().plus(maxWaitUntilReady);
    }

    private void registerStop() {
        timeoutTime = null;
    }

    @Nonnull
    @Override
    protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {

        registerStart();


        if (waitForFirstCheck)
            waitForCheck();

        boolean finalState = false;
        boolean fatalStop = false;
        while (!finalState && !fatalStop && !isTimeout()) {
            //4. Wait for the resources to become ready
            experimentPart.setState(InternalState.WAIT_FOR_READY);

            boolean statusIsFinished = false;

            while (!statusIsFinished && !fatalStop && !isTimeout()) {

                //Check status, and keep user updated on the progress
                CountDownLatch statusTaskExecutionReadyLatch = new CountDownLatch(1);

                this.job.submitTask(statusTask, (TaskExecutionFinishedCallback) (task, taskExecution, state) -> statusTaskExecutionReadyLatch.countDown());
                triesCount++;

                int punctuationCount = -1;

                do {
                    updateMessage(String.format("Checked status %d times. Checking%s",
                            triesCount,
                            Strings.repeat(".", 1 + ++punctuationCount % 3)));
                } while (!statusTaskExecutionReadyLatch.await(1, TimeUnit.SECONDS));


                /* Check all factors that can influence or decision to keep requesting the status.
                If one of these factors is not OK anymore, set fatalStop to TRUE to halt the process. */

                if (statusTask.isBusy()) {
                    busyCount++;

                    if (busyCount > MAX_BUSY_COUNT) {
                        LOG.warn("Got BUSY too many times for resources on {} of {}", experimentPart.getName(), experiment.getName());
                        updateMessage("Waited too long for resources on '" + experimentPart.getName() + "' to become ready." +
                                " (Waited " + busyCount + " times 10 seconds.)");
                        fatalStop = true;
                    }
                }

                if (statusTask.isHardFailure())
                    fatalStop = true;
                else if (statusTask.isSoftFailure()) {
                    softFailCount++;

                    if (softFailCount > MAX_SOFT_FAIL_COUNT) {
                        LOG.warn("Something went wrong while waiting for resources on '\"+experimentPart.getName()+\"' to become ready." +
                                " Will stop checking status...");
                        updateMessage("Fetching status failed. Aborting checks!");

                        fatalStop = true;
                    }
                } else {
                    assert statusTask.getStatusDetails() != null;

                    if (statusTask.getStatusDetails() != null &&
                            hasAtLeastOneFailure(statusTask.getStatusDetails())) {
                        resourcesFailedRetryCount++;

                        if (resourcesFailedRetryCount > MAX_RESOURCES_FAILED_RETRY_COUNT) {
                            LOG.warn("At least one resource has FAILED  on '{}'. Retried {} times, but there's no point to keep retrying." +
                                    " Will stop checking status...", experimentPart.getName(), MAX_RESOURCES_FAILED_RETRY_COUNT);
                            fatalStop = true;

                            updateMessage("One or more resources have permanently failed");

                        }
                    } else {
                        resourcesFailedRetryCount = 0;
                    }
                }

                statusIsFinished = isStatusTaskFinished();

                if (!statusIsFinished)
                    waitForCheck();
            }

            //log why we stopped listening
            LOG.info("Stopped checking status for {} because {}.",
                    experimentPart.getName(),
                    statusIsFinished ? "statusIsFinalState"
                            : (fatalStop ? "fatalStop" : (isTimeout() ? "timeout" : "UNKNOWN")));

            hadTimeout = isTimeout();

            if (statusIsFinished && desiredStatus == DesiredStatus.GENI_READY) {
                LOG.debug("Stopped listening for results of status calls after receiving status {} ",
                        statusTask.getStatusDetails() != null ? statusTask.getStatusDetails().getGlobalStatus()
                                : "null");

                //5. Do a list-resources to verify what we got
                ListSliceResourcesTask sliceResourcesTask =
                        hltf.listSliceResources(experiment.getSlice(), experimentPart.getConnectSfaAuthority());

                CountDownLatch listResourcesTaskReadyLatch = new CountDownLatch(1);
                int punctuationCount = -1;

                this.job.submitTask(sliceResourcesTask, (TaskExecutionFinishedCallback) (task, taskExecution, state) -> listResourcesTaskReadyLatch.countDown());

                do {
                    updateMessage(String.format("Verifying final state%s",
                            Strings.repeat(".", 1 + ++punctuationCount % 3)));
                } while (!listResourcesTaskReadyLatch.await(1, TimeUnit.SECONDS));
                updateMessage("");

                finalState = areAllResourcesInDesiredState(DesiredStatus.GENI_READY);

                if (sliceResourcesTask.isHardFailure()) {
                    updateMessage("Something went wrong with the ListResource call. Aborting periodic check.");
                    fatalStop = true;
                }

                if (!finalState) {
                    LOG.warn("ListResources reported that not all resources from {} are in a final state. Will start checking again.", experimentPart.getName());
                }
            }
        }
        if (isTimeout())
            hadTimeout = true;
        registerStop();

        if (finalState) {
            if (statusTask.getStatusDetails() == null) {
                LOG.error("statusTask.getStatusDetails() == null is unexpected. Considering this a FAILURE.");
                return ExperimentTaskStatus.FAILED;
            }
            if (hasAtLeastOneFailure(statusTask.getStatusDetails()) ||
                    statusTask.getStatusDetails().getGlobalStatus() == StatusDetails.SliverStatus.FAIL) {
                return ExperimentTaskStatus.FAILED;
            } else {
                //Check if the non-fail final state, is the state we wanted to see.
                return areAllResourcesInDesiredState(desiredStatus) ?
                        ExperimentTaskStatus.SUCCESS :
                        ExperimentTaskStatus.FAILED;
            }
        } else {
            return ExperimentTaskStatus.FAILED;
        }
    }

    public boolean isWaitForFirstCheck() {
        return waitForFirstCheck;
    }

    public void setWaitForFirstCheck(boolean waitForFirstCheck) {
        this.waitForFirstCheck = waitForFirstCheck;
    }

    private void waitForCheck() throws InterruptedException {
        //Wait the requested time before querying status
        long checkReadyIntervalMs = checkReadyInterval.toMillis();
        int waitedMs = 0;
        while (waitedMs < checkReadyIntervalMs) {
            updateMessage(String.format("Checked status %d times. Next check in %d seconds",
                    triesCount,
                    roundUp((checkReadyIntervalMs - waitedMs), 1000)));

            Thread.sleep(1000);
            waitedMs += 1000;
        }
    }

    private boolean isStatusTaskFinished() {
        if (statusTask.isBusy())
            return false;

        if (statusTask.getStatusDetails() == null)
            return false;


        //explicitly check all sub-statuses, because the general status is not always correct (FEDIBBTDEV-2178)
        for (StatusDetails.SliverStatus sliverStatus :
                Iterables.concat(statusTask.getStatusDetails().getStatusBySliverUrn().values(),
                        statusTask.getStatusDetails().getStatusByComponentUrn().values())) {
            if (sliverStatus == StatusDetails.SliverStatus.UNKNOWN ||
                    sliverStatus == StatusDetails.SliverStatus.CHANGING)
                return false;
        }

        if (statusTask.getStatusDetails().getGlobalStatus() == StatusDetails.SliverStatus.UNKNOWN)
            return false;
        if (statusTask.getStatusDetails().getGlobalStatus() == StatusDetails.SliverStatus.CHANGING)
            return false;

        return true;
    }


    private static boolean hasAtLeastOneFailure(@Nonnull StatusDetails statusDetails) {
        for (StatusDetails.SliverStatus sliverStatus : statusDetails.getStatusBySliverUrn().values())
            if (sliverStatus == StatusDetails.SliverStatus.FAIL)
                return true;
        for (StatusDetails.SliverStatus sliverStatus : statusDetails.getStatusByComponentUrn().values())
            if (sliverStatus == StatusDetails.SliverStatus.FAIL)
                return true;

        return false;
    }


    private boolean areAllResourcesInDesiredState(@Nonnull DesiredStatus desiredStatus) {
        assert experiment.getSlice() != null;
        List<Sliver> sliversOfAuthority = experiment.getSlice().findSlivers(experimentPart.getConnectSfaAuthority());


        for (Sliver sliver : sliversOfAuthority) {
            if (!sliver.isFake() && sliver.getNodes() == null) {
                LOG.debug("No nodes available in sliver {}", sliver.getUrn());
                continue;
            }

            StatusDetails statusDetails = experiment.getSlice().getStatus();
            if (statusDetails == null) {
                LOG.error("No status info available! Fallback: assuming resources are NOT in the required state.");
                return false;
            }

            //treat fake slivers differently, they don't really have theri own status. Their status is the same as the global status.
            if (sliver.isFake()) {
                if (experiment.getSlice().getStatus().getGlobalStatus() == null) {
                    LOG.error("No global status info available! Fallback: assuming fake resource(s) are NOT in the required state.");
                    return false;
                }
                switch (experiment.getSlice().getStatus().getGlobalStatus()) {
                    case UNINITIALISED: //fall-through
                    case UNALLOCATED:   //fall-through
                    case UNKNOWN:       //fall-through
                    case CHANGING:      //fall-through
                    case FAIL:
                        return false;
                    case READY: {
                        if (desiredStatus != DesiredStatus.GENI_READY)
                            return false;
                        break;
                    }
                    case NOTREADY: {
                        if (desiredStatus != DesiredStatus.GENI_NOTREADY)
                            return false;
                        break;
                    }
                }
                continue;
            }

            final StatusDetails.SliverStatus statusBySliverUrn =
                    statusDetails.getStatusBySliverUrn(sliver.getUrn());
            if(statusBySliverUrn == StatusDetails.SliverStatus.CHANGING ||
                    statusBySliverUrn == StatusDetails.SliverStatus.UNKNOWN)
                return false;

            if (desiredStatus == DesiredStatus.GENI_NOTREADY &&
                    statusBySliverUrn != StatusDetails.SliverStatus.NOTREADY)
                return false;
            if (desiredStatus == DesiredStatus.GENI_READY &&
                    statusBySliverUrn != StatusDetails.SliverStatus.READY)
                return false;

            if (sliver.getNodes() != null) {
                for (RspecNode rspecNode : sliver.getNodes()) {
                    if (experiment.getSlice().getStatus().getStatusByComponentUrn().containsKey(rspecNode.getComponentId())) {
                        StatusDetails.SliverStatus nodeStatus =
                                experiment.getSlice().getStatus().getStatusByComponentUrn().get(rspecNode.getComponentId());
                        if (nodeStatus == StatusDetails.SliverStatus.CHANGING
                                || nodeStatus == StatusDetails.SliverStatus.UNKNOWN) {
                            LOG.debug("rspecNode {} not in state {} but in {}",
                                    rspecNode.getClientId(), desiredStatus, nodeStatus);
                            return false;
                        }

                        if (desiredStatus == DesiredStatus.GENI_NOTREADY &&
                                nodeStatus != StatusDetails.SliverStatus.NOTREADY) {
                            LOG.debug("rspecNode {} not in state {} but in {}",
                                    rspecNode.getClientId(), desiredStatus, nodeStatus);
                            return false;
                        }
                        if (desiredStatus == DesiredStatus.GENI_READY &&
                                nodeStatus != StatusDetails.SliverStatus.READY) {
                            LOG.debug("rspecNode {} not in state {} but in {}",
                                    rspecNode.getClientId(), desiredStatus, nodeStatus);
                            return false;
                        }
                    } else
                        LOG.debug("No status by componentId for {}", rspecNode.getComponentId());
                }
            }
        }
        return true;

    }

    private boolean isTimeout() {
        return timeoutTime != null && timeoutTime.isBefore(Instant.now());
    }

    public boolean isFailedDueToTimeout() {
        return hadTimeout;
    }

    private static long roundUp(long num, long divisor) {
        return (num + divisor - 1) / divisor;
    }

    public void setCheckReadyIntervalMs(long checkReadyIntervalMs) {
        this.checkReadyInterval = Duration.ofMillis(checkReadyIntervalMs);
    }
    public void setCheckReadyInterval(@Nullable Duration checkReadyInterval) {
        this.checkReadyInterval =
                checkReadyInterval == null ? DEFAULT_CHECK_READY_INTERVAL : checkReadyInterval;
    }
}
